/* Experimenteller Webserver 0.2 (c) 2003 by Dennis Schwerdel
 * Dieses Programm ist freie Software, es kann nur im Rahmen der
 * GNU General Public License (GPL) Version 2 oder neuer verwendet werden.
 * -> Siehe license.txt
 */

package server.httpd ;

import server.RequestThread ;

import java.net.Socket;
import java.net.URL ;
import java.net.URLConnection ;
import java.io.PrintStream;
import java.io.InputStream ;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.File;
import java.util.StringTokenizer;
import java.util.Date;

public class HttpRequestThread extends RequestThread {

  protected HttpSettings settings ;
  protected String headers = "" ;
  protected String buffer ;
  protected boolean headers_written = false ;

  public HttpRequestThread ( HttpSettings settings ) {
    super () ;
    this.settings = settings ;
  }

  public HttpRequestThread ( Socket socket, HttpSettings settings ) {
    super ( socket ) ;
    this.settings = settings ;
  }

  /********************************************************************************************************
   ********************************************************************************************************
   ********************************************************************************************************/

  protected void addHeader ( String header ) {
    if ( ! headers_written ) headers += header + "\r\n" ;
  }

  protected void writeHeaders () {
    if ( headers_written ) return ;
    headers_written = true ;
    super.write ( headers ) ;
    super.write ( "\r\n" ) ;
    if ( buffer != null ) {
      super.write ( buffer ) ;
      buffer = null ;
    }
  }

  protected void write ( String str ) {
    if ( ! headers_written ) {
      buffer += str ;
    } else super.write ( str ) ;
  }

  /********************************************************************************************************
   ********************************************************************************************************
   ********************************************************************************************************/

  public void handleRequest ( String request ) throws Exception {
    if ( request.indexOf ( "GET" ) >= 0 ) {
      StringTokenizer lines = new StringTokenizer ( request, "\n" ) ;
      String params = "" ;
      String file = "" ;
      int from = 0 ;
      int to = -1 ;
      while ( lines.hasMoreTokens () ) {
        String line = lines.nextToken () ;
        StringTokenizer st = new StringTokenizer ( line ) ;
        String command = st.nextToken () ;
	if ( command.toUpperCase().equals( "GET" ) ) {
          String get = st.nextToken () ;
          int sep = get.indexOf ( '?' ) ;
          if ( sep >= 0 ) {
            file = get.substring ( 0, sep ) ;
	    params = get.substring ( sep + 1 ) ;
	  } else file = get ;
	}
	if ( command.toUpperCase().equals( "RANGE:" ) ) {
	  String part2 = st.nextToken () ;
	  part2 = part2.substring ( part2.indexOf ( "=" ) + 1 ) ;
	  from = Integer.parseInt ( ( "0" + part2.substring ( 0, part2.indexOf ( "-" ) ) ).trim () ) ;
	  to = Integer.parseInt ( ( "0" + part2.substring ( part2.indexOf ( "-" ) + 1 ) ).trim () ) ;
	  if ( to <= from ) to = -1 ;
	}
      }
      internal_request ( file, params, from, to ) ;
    }
  }

  protected void internal_request ( String file, String params, int from, int to ) throws Exception {
    request_file ( file, from, to ) ;
  }

  protected void error ( int code, String desc ) {
    switch ( code ) {
      default : addHeader ( "HTTP/1.0 " + code + " " + desc ) ;
                writeHeaders () ;
                write ( "<html><head><title>" + desc + "</title></head>" ) ;
	        write ( "<body><h1>HTTP Error " + code + "</h1><h2>" + desc + "</h2></body></html>" ) ;
    }
    System.err.println ( "HTTP Error " + code + " : " + desc ) ;
  }

  protected void printDataHeader ( int length, int from, int to, String mime ) {
    if ( length != to - from + 1 ) {
      addHeader ( "HTTP/1.0 206 Partial content" ) ;
    } else {
      addHeader ( "HTTP/1.0 200 OK" ) ;
    }
    addHeader ( "Content-type: " + mime ) ;
    addHeader ( "Content-length: " + ( to - from + 1 ) ) ;
    addHeader ( "Content-range: bytes " + from + "-" + to + "/" + length ) ;
    writeHeaders () ;
  }
  
  protected void request_file ( String virtualFile, int from, int to ) throws Exception {
    settings.logfile.write ( ( new Date ().toString() + " " + socket.getInetAddress().getHostAddress() + " GET " + virtualFile + "\n" ).getBytes () ) ;
    try {
      String realFile = settings.paths.resolve ( virtualFile ) ;
      String mimeType = settings.mimes.resolve ( realFile ) ;
      String exec = settings.mimes.exec ( realFile ) ;
      File fileHandle = new File ( realFile ) ;
      int length = 0 ;
      if ( ! fileHandle.exists() ) {
        // Datei existiert nicht, vom Mirror nachladen
        String mirrorURL = settings.paths.mirror ( virtualFile ) ;
        if ( mirrorURL == null || mirrorURL == "" ) throw new FileNotFoundException ( "File not found : " + virtualFile ) ;
	URLConnection mirrorConnection = new URL ( mirrorURL ).openConnection() ;	
	length = (int)mirrorConnection.getContentLength() ;
	if ( to < 1 ) to = length - 1 ;
        to = Math.max ( 0, Math.min ( length - 1, to ) ) ;
	from = Math.max ( 0, Math.min ( to, from ) ) ;
        printDataHeader ( length, from, to, mimeType ) ;
	printGetFile ( mirrorConnection, fileHandle, settings.buffersize, from, to ) ;
      } else if ( fileHandle.isDirectory () ) {
        // Datei ist ein Verzeichnis
        addHeader ( "HTTP/1.0 200 OK" ) ;
        dirlist ( fileHandle, virtualFile ) ;
      } else if ( exec == null || exec.equals ( "" ) ) {
        // ganz normale Datei
	length = (int)fileHandle.length() ;
	if ( to < 1 ) to = length - 1 ;
        to = Math.max ( 0, Math.min ( length - 1, to ) ) ;
	from = Math.max ( 0, Math.min ( to, from ) ) ;
        printDataHeader ( length, from, to, mimeType ) ;
	printFile ( fileHandle, settings.buffersize, from, to ) ;
      } else {
        // Datei muss ausgefhrt werden
	InputStream data = settings.exec.run ( virtualFile, realFile ) ;
	length = data.available() ;
	if ( to < 1 ) to = length - 1 ;
        to = Math.max ( 0, Math.min ( length - 1, to ) ) ;
	from = Math.max ( 0, Math.min ( to, from ) ) ;
        printDataHeader ( length, from, to, mimeType ) ;
	printStream ( data, settings.buffersize, from, to ) ;
      }
    } catch ( FileNotFoundException e ) {
      if ( e.getMessage ().indexOf ( "Permission denied" ) >= 0 ) {
        error ( 403, "Permission denied : " + virtualFile ) ;
      } else {
        error ( 404, "File Not Found : " + virtualFile ) ;
      }
    } catch ( Exception e ) {
      error ( 500, "Internal Server Error : " + e.toString () ) ;
    }
  }
      
  protected void dirlist ( File theFile, String base ) throws Exception {
    String [] dir = settings.paths.dirlist ( base ) ;
    if ( base.lastIndexOf ( "/" ) == base.length () - 1 ) base = base.substring ( 0, base.length () - 1 ) ;
    addHeader ( "Content-type: text/html" ) ;
    writeHeaders () ;
    write ( "<HTML><Body><Center><H1>Index of " + base + "</H1>" ) ;
    write ( "<Table><TR><TD>Name</TD><TD>Size</TD></TR>\r\n" ) ;
    write ( "<TR><TD><A Href='" + base + "/..'>-[ .. ]-</A></TD><TD></TD></TR>\r\n" ) ;
    for ( int i = 0; i < dir.length; i++ ) {
      File item = new File ( theFile.getCanonicalPath() + "/" + dir[i] ) ;
      double size = item.length() ;
      size = Math.round ( (float)(size / 102.4) ) / 10.0 ;
      String sizeStr = "k" ;
      if ( size >= 1024 ) {
        size = Math.round ( (float)(size / 102.4) ) / 10.0 ;
        sizeStr = "M" ;
      }
      if ( size >= 1024 ) {
        size = Math.round ( (float)(size / 102.4) ) / 10.0 ;
        sizeStr = "G" ;
      }
      sizeStr = size + " " + sizeStr ;
      String file = dir[i] ;
      if ( file.substring ( file.length() - 1 ).equals ( "-" ) ) file = file.substring ( 3, file.length() - 3 ) ;
      write ( "<TR><TD><A Href='" + base + "/" + file + "'>" + dir[i] + "</A></TD><TD align=right>" + sizeStr + "</TD></TR>\r\n" ) ;
    }
    write ( "</Table></Center></Body></HTML>" ) ;
  }

}
